using System;
using System.Collections.Generic;
using System.Linq;
using Clarion;
using Clarion.Framework;
using Clarion.Framework.Core;
using Clarion.Framework.Templates;
using WorldServerLibrary.Model;
using System.Threading;

namespace WorldServerAgent
{
	 /// <summary>

    /// This event will be fired at the beginning of each cognitive cycle in order to get the current visual sensor information.
    /// </summary>
    /// <returns>The Visual Sensor Information Itens</returns> 
    public delegate IList<Thing> InputVisualSensorialInformationEventHandler();

    /// <summary>
    /// This event will be fired at the end of each cognitive cycle in order to translate the selected action to a World Server Command
    /// </summary>
    /// <param name="externalAction">The external action that was choice</param>
    public delegate void OutputActionChunkEventHandler(ClarionAgentActionType externalAction);

    /// <summary>
    /// Public enum that represents all possibilities of agent actions
    /// </summary>
    public enum ClarionAgentActionType
    {
        DO_NOTHING,
        ROTATE_CLOCKWISE,
        GO_AHEAD,
		GO_TO_JEWEL,
		GO_TO_FOOD,
		SACK_IT		
    }
	
			
    public class ClarionAgent
    {
        #region Constants
		
		/// <summary>
    	/// Parameter list
    	/// </summary>
    	public double[] param = {0,0,0,0};
		public double[] param_food = {0,0,0,0};		
		public string  foodName = "";
		public Thing SackItItem;
		
		public Creature Oss = new Creature();
		
        /// <summary>
        /// Constant that represents the Visual Sensor
        /// </summary>
        private String SENSOR_VISUAL_DIMENSION = "VisualSensor";		

        /// <summary>
        /// Constant that represents that there is at least one wall ahead
        /// </summary>
        private String DIMENSION_WALL_AHEAD = "WallAhead";
		private int JEWEL_DETECTED = 0;
		private String DIMENSION_FOOD_AHEAD = "FoodAhead";
		private String DIMENSION_SACK_IT = "SackIt";		
		
        #endregion

        #region Properties

        #region Simulation

        /// <summary>
        /// If this value is greater than zero, the agent will have a finite number of cognitive cycle. Otherwise, it will have infinite cycles.
        /// </summary>
        public double MaxNumberOfCognitiveCycles = -1;

        /// <summary>
        /// Current cognitive cycle number
        /// </summary>
        private double CurrentCognitiveCycle = 0;

        /// <summary>
        /// Time between cognitive cycle in miliseconds
        /// </summary>
        public Int32 TimeBetweenCognitiveCycles = 0;

        /// <summary>
        /// A thread Class that will handle the simulation process
        /// </summary>
        private Thread runThread;

        #endregion

        #region Agent

        /// <summary>
        /// The agent 
        /// </summary>
        private Clarion.Framework.Agent CurrentAgent;

        /// <summary>
        /// For each cognitive cycle, this event will be called in order to the agent receives the current sensorial information
        /// </summary>
        public event InputVisualSensorialInformationEventHandler OnNewVisualSensorialInformation;

        /// <summary>
        /// For each cognitive cycle, this event will be called when the agent selects one action
        /// </summary>
        public event OutputActionChunkEventHandler OnNewExternalActionSelected;

        #endregion

        #region Perception Input

        /// <summary>
        /// Perception input to indicates a wall ahead
        /// </summary>
        private DimensionValuePair InputWallAhead;
		private DimensionValuePair InputJewel;
		private DimensionValuePair InputFoodAhead;
		private DimensionValuePair InputSackIt;	

        #endregion

        #region Action Output

        /// <summary>
        /// Output action that makes the agent to rotate clockwise
        /// </summary>
        private ExternalActionChunk OutputRotateClockwise;

        /// <summary>
        /// Output action that makes the agent go ahead
        /// </summary>
        private ExternalActionChunk OutputGoAhead;
		
		/// <summary>
        /// Output action that makes the agent go to the Jewel
        /// </summary>
		private ExternalActionChunk OutputGoToJewel;
		
		/// <summary>
        /// Output action that makes the agent go to the Jewel
        /// </summary>
		private ExternalActionChunk OutputGoToFood;
		
		/// <summary>
		/// The output action sack it.
		/// </summary>
		private ExternalActionChunk OutputSackIt;
		
        #endregion

        #endregion

        #region Constructor

        public ClarionAgent()
        {
            // Initialize the agent
            CurrentAgent = World.NewAgent("Current Agent");

            // Initialize Input Information
            InputWallAhead = World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, DIMENSION_WALL_AHEAD);			
			InputJewel = World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, JEWEL_DETECTED);
			InputFoodAhead = World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, DIMENSION_FOOD_AHEAD);
			InputSackIt = World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, DIMENSION_SACK_IT);			
			
            // Initialize Output actions
            OutputRotateClockwise = World.NewExternalActionChunk(ClarionAgentActionType.ROTATE_CLOCKWISE.ToString());
            OutputGoAhead = World.NewExternalActionChunk(ClarionAgentActionType.GO_AHEAD.ToString());
			OutputGoToJewel = World.NewExternalActionChunk(ClarionAgentActionType.GO_TO_JEWEL.ToString());
			OutputGoToFood = World.NewExternalActionChunk(ClarionAgentActionType.GO_TO_FOOD.ToString());
			OutputSackIt = World.NewExternalActionChunk(ClarionAgentActionType.SACK_IT.ToString());
						
            //Create thread to simulation
            runThread = new Thread(RunThread);
			Console.WriteLine("Agent started");
        }

        #endregion

        #region Public Methods

        /// <summary>
        /// Run the Simulation in World Server 3d Environment
        /// </summary>
        public void Run()
        {                
			Console.WriteLine ("Running ...");
            // Setup Agent to run
            if (runThread != null && !runThread.IsAlive)
            {
                SetupAgentInfraStructure();

                if (OnNewVisualSensorialInformation != null)
                {
                    // Start Simulation Thread                
                    runThread.Start(null);
                }
            }
        }

        /// <summary>
        /// Abort the current Simulation
        /// </summary>
        /// <param name="deleteAgent">If true beyond abort the current simulation it will die the agent.</param>
        public void Abort(Boolean deleteAgent)
        {   Console.WriteLine ("Aborting ...");
            if (runThread != null && runThread.IsAlive)
            {
                runThread.Abort();
            }

            if (CurrentAgent != null && deleteAgent)
            {
                CurrentAgent.Die();
            }
        }

        #endregion

        #region Setup Agent Methods

        /// <summary>
        /// Setup agent infra structure (ACS, NACS, MS and MCS)
        /// </summary>
        private void SetupAgentInfraStructure()
        {
            // Setup the ACS Subsystem
            SetupACS();                    
        }

        private void SetupMS()
        {            
            //RichDrive
        }

        /// <summary>
        /// Setup the ACS subsystem
        /// </summary>
        private void SetupACS()
        {
            // Create Colission Wall Rule
            SupportCalculator avoidCollisionWallSupportCalculator = FixedRuleDelegateToAvoidColissionWall;
            FixedRule ruleAvoidCollisionWall = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputRotateClockwise, avoidCollisionWallSupportCalculator);

            // Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit(ruleAvoidCollisionWall);

            // Create Go To Ahead
            SupportCalculator goAheadSupportCalculator = FixedRuleDelegateToGoAhead;
            FixedRule ruleGoAhead = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputGoAhead, goAheadSupportCalculator);
            
            // Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit(ruleGoAhead);
			
			
			// Create GoToJewel Rule
			SupportCalculator goToJewelSupportCalculator = FixedRuleDelegateToGoToJewel;
			FixedRule ruleGoToJewel = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputGoToJewel, goToJewelSupportCalculator);
			
			// Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit(ruleGoToJewel);
			
					
			// Create GoToFood Rule
			SupportCalculator goToFoodSupportCalculator = FixedRuleDelegateToGoToFood;			
			FixedRule ruleGoToFood = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputGoToFood, goToFoodSupportCalculator);
			
			// Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit(ruleGoToFood);
			
			// Create GoToFood SackIt
			SupportCalculator SackItCalculator = FixedRuleDelegateSackIt;
			FixedRule ruleSackIt = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputSackIt, SackItCalculator);
			
			// Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit(ruleSackIt);		
			
			
            // Disable Rule Refinement
            CurrentAgent.ACS.Parameters.PERFORM_RER_REFINEMENT = false;

            // The selectio type will be probabilistic
            CurrentAgent.ACS.Parameters.LEVEL_SELECTION_METHOD = ActionCenteredSubsystem.LevelSelectionMethods.STOCHASTIC;

            // The selection of the action will be fixed (not variable) i.e. only the statement defined above.
            CurrentAgent.ACS.Parameters.LEVEL_SELECTION_OPTION = ActionCenteredSubsystem.LevelSelectionOptions.FIXED;
				

            // Define Probabilistic values
			CurrentAgent.ACS.Parameters.FIXED_FR_LEVEL_SELECTION_MEASURE = 1;			
            CurrentAgent.ACS.Parameters.FIXED_IRL_LEVEL_SELECTION_MEASURE = 0;
            CurrentAgent.ACS.Parameters.FIXED_BL_LEVEL_SELECTION_MEASURE = 0;
            CurrentAgent.ACS.Parameters.FIXED_RER_LEVEL_SELECTION_MEASURE = 0;
			
        }

        /// <summary>
        /// Make the agent perception. In other words, translate the information that came from sensors to a new type that the agent can understand
        /// </summary>
        /// <param name="sensorialInformation">The information that came from server</param>
        /// <returns>The perceived information</returns>
        private SensoryInformation MakePerceptionFromSensorialInput(IList<Thing> sensorialInformation)
        {
            // New sensory information
            SensoryInformation si = World.NewSensoryInformation(CurrentAgent);
					
			// Detect if we have a wall ahead
            Boolean wallAhead = sensorialInformation.Where(item => (item.CategoryId == Thing.CATEGORY_BRICK && item.DistanceToCreature <= 35)).Any();
            
			if (wallAhead)
			{				
				si.Add(InputSackIt, CurrentAgent.Parameters.MIN_ACTIVATION);
				si.Add(InputFoodAhead, CurrentAgent.Parameters.MIN_ACTIVATION);
				si.Add(InputJewel, CurrentAgent.Parameters.MIN_ACTIVATION);
				si.Add(InputWallAhead, CurrentAgent.Parameters.MAX_ACTIVATION);
				
				return si;
			}
			
						
			List<Thing> JewelList = new List<Thing>();
			List<Thing> FoodList = new List<Thing>();
			
			foreach (Thing item in sensorialInformation)
			{
				// Detect if there is something to sack
				if(item.CategoryId != Thing.CATEGORY_CREATURE && item.CategoryId != Thing.CATEGORY_PFOOD &&  item.CategoryId != Thing.CATEGORY_NPFOOD && item.CategoryId != Thing.CATEGORY_BRICK && item.DistanceToCreature <= 40)
				{	
					SackItItem = item;
					si.Add(InputSackIt, CurrentAgent.Parameters.MAX_ACTIVATION);
					si.Add(InputFoodAhead, CurrentAgent.Parameters.MIN_ACTIVATION);
					si.Add(InputJewel, CurrentAgent.Parameters.MIN_ACTIVATION);					
					return si;
				}
				
				// If there is a food 
				if(item.CategoryId == Thing.CATEGORY_PFOOD && item.DistanceToCreature <= 40 && Oss.Fuel >= 500)
				{	
					si.Add(InputWallAhead, CurrentAgent.Parameters.MAX_ACTIVATION);
					si.Add(InputSackIt, CurrentAgent.Parameters.MIN_ACTIVATION);
					si.Add(InputJewel, CurrentAgent.Parameters.MIN_ACTIVATION);
					si.Add(InputFoodAhead, CurrentAgent.Parameters.MIN_ACTIVATION);
					return si;
				}							
				
				// Detect if we have a Food
				if(item.CategoryId == Thing.CATEGORY_PFOOD)
					FoodList.Add(item);
								
				// Detect if we have a Jewel
				if(item.CategoryId == Thing.CATEGORY_JEWEL)
					JewelList.Add(item);				
			}
								
			
			if(Oss.Fuel < 500)
			{
				if(FoodList.Count > 0)
				{
					Console.Out.WriteLine("Baja energia \n");
				
					FoodList = FoodList.OrderBy(o=>o.DistanceToCreature).ToList();
				
					int x =  (int)FoodList[0].X1;
					int y = (int)FoodList[0].Y1;
					int dis = (int)FoodList[0].DistanceToCreature;
				
								
					param_food[0] = x;
					param_food[1] = y;
					param_food[2] = dis;						
					foodName = FoodList[0].Name;
				
					//Console.WriteLine("Comida!!!, Pos: " + x.ToString() + " , " + y.ToString() + " Dist = " + dis.ToString());

					si.Add(InputWallAhead, CurrentAgent.Parameters.MIN_ACTIVATION);
					si.Add(InputSackIt, CurrentAgent.Parameters.MIN_ACTIVATION);
					si.Add(InputJewel, CurrentAgent.Parameters.MIN_ACTIVATION);			
					si.Add(InputFoodAhead, CurrentAgent.Parameters.MAX_ACTIVATION);			
					
					return si;
				}
				
			}
			
			if(JewelList.Count > 0)
			{
				//Console.WriteLine("Found " + JewelList.Count.ToString() + " Jewels !!");	
				
				
				JewelList = JewelList.OrderBy(o=>o.DistanceToCreature).ToList();
				
				int x =  (int)JewelList[0].X1;
				int y = (int)JewelList[0].Y1;
				int dis = (int)JewelList[0].DistanceToCreature;
				param[0] = x;
				param[1] = y;
				param[2] = dis;
				
				si.Add(InputJewel, CurrentAgent.Parameters.MAX_ACTIVATION);
				
				return si;
			}
			
			
			si.Add(InputWallAhead, CurrentAgent.Parameters.MAX_ACTIVATION);
			si.Add(InputSackIt, CurrentAgent.Parameters.MIN_ACTIVATION);
			si.Add(InputJewel, CurrentAgent.Parameters.MIN_ACTIVATION);			
			si.Add(InputFoodAhead, CurrentAgent.Parameters.MIN_ACTIVATION);			
			
            return si;
        }

        #endregion

        #region Delegate Methods

        #region Fixed Rules

        private double FixedRuleDelegateToAvoidColissionWall(ActivationCollection currentInput, Rule target)
        {			
            return ((currentInput.Contains(InputWallAhead, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
        }
		
		private double FixedRuleDelegateSackIt(ActivationCollection currentInput, Rule target)
        {	
			if(currentInput.Contains(InputWallAhead, CurrentAgent.Parameters.MAX_ACTIVATION))
			   return 0;
			
			if((currentInput.Contains(InputSackIt, CurrentAgent.Parameters.MAX_ACTIVATION)))
				return 1;
			
			return 0;
		}
		
        private double FixedRuleDelegateToGoAhead(ActivationCollection currentInput, Rule target)
        {	
			if(currentInput.Contains(InputWallAhead, CurrentAgent.Parameters.MIN_ACTIVATION) && currentInput.Contains(InputFoodAhead, CurrentAgent.Parameters.MIN_ACTIVATION) && currentInput.Contains(InputJewel, CurrentAgent.Parameters.MIN_ACTIVATION))
				return 1;
			else
				return 0;
			
        }
		
		private double FixedRuleDelegateToGoToFood(ActivationCollection currentInput, Rule target)
        {
			if(currentInput.Contains(InputWallAhead, CurrentAgent.Parameters.MAX_ACTIVATION))
				return 0;
			
			if(currentInput.Contains(InputFoodAhead, CurrentAgent.Parameters.MAX_ACTIVATION))			
				return 1;
			
			return 0;	
        }
		
		private double FixedRuleDelegateToGoToJewel(ActivationCollection currentInput, Rule target)
        {
			if(currentInput.Contains(InputWallAhead, CurrentAgent.Parameters.MAX_ACTIVATION))
			   return 0;
			
			if(currentInput.Contains(InputSackIt, CurrentAgent.Parameters.MAX_ACTIVATION))
			   return 0;
					
			if(currentInput.Contains(InputFoodAhead, CurrentAgent.Parameters.MAX_ACTIVATION))
			   return 0;
			
			if(currentInput.Contains(InputJewel , CurrentAgent.Parameters.MAX_ACTIVATION))
			   return 1;
			
			return 0;
			
        }
		
		

        #endregion

        #region Run Thread Method

        private void RunThread(object obj)
        {
			Console.WriteLine("Starting Cognitive Cycle ... press CTRL-C to finish !");
            // Cognitive Cycle starts here getting sensorial information
            while (CurrentCognitiveCycle != MaxNumberOfCognitiveCycles)
            {   
                // Get current sensorial information                    
                IList<Thing> sensorialInformation = OnNewVisualSensorialInformation();

                // Make the perception
                SensoryInformation si = MakePerceptionFromSensorialInput(sensorialInformation);

                //Perceive the sensory information
                CurrentAgent.Perceive(si);

                //Choose an action
                ExternalActionChunk chosen = CurrentAgent.GetChosenExternalAction(si);

                // Get the selected action
                String actionLabel = chosen.LabelAsIComparable.ToString();
				
				//Console.Out.WriteLine(actionLabel + "\n");
				
                ClarionAgentActionType actionType = (ClarionAgentActionType)Enum.Parse(typeof(ClarionAgentActionType), actionLabel, true);

                // Call the output event handler
                if (OnNewExternalActionSelected != null)
                {
                    OnNewExternalActionSelected(actionType);
                }

                // Increment the number of cognitive cycles
                CurrentCognitiveCycle++;

                //Wait to the agent accomplish his job
                if (TimeBetweenCognitiveCycles > 0)
                {
                    Thread.Sleep(TimeBetweenCognitiveCycles);
                }
            }
        }

        #endregion

        #endregion
    }
}

